#!/usr/bin/env python

import sys

def usage():
    print '''%s [input_fname] [#(get)] [#(put)]
 '''  %(sys.argv[0])
 
def is_var_bb(var):
	return var[0:2] == "BB"

def is_var_cs(var):
	return var[0:2] == "CS"

def is_var_get(var):
	return var[3:4] == "G"

def is_var_put(var):
	return var[3:4] == "P"

def is_var_idle(var):
	return var[0] == "I"


def subseq_to_sumexp(subseq):	
	output_exp = ""
	subseq0 = subseq[0].replace("\n", "")
	if len(subseq) == 1:
		output_exp = subseq0
	else:
		output_exp = "(+ %s %s)" %(subseq0, subseq_to_sumexp(subseq[1:]))
	return output_exp

'''
@op: string, '<', '=', or '>'
'''
def subseq_sum_assert(subseq, op, assertval):
	out_exp = subseq_to_sumexp(subseq)
	out_exp = "(assert(%s %s %d))" %(op, out_exp, assertval)
	return out_exp
	
def process_subseq(subseq):
	sys.stderr.write("processing busy periods (%d records)\n" %len(subseq))
	out_exp = ""
	for i in xrange(0, len(subseq)):
		var = subseq[i]
		if is_var_bb(var) and is_var_get(var):
			# a new basic block. refcount must > 0 
			out_exp += subseq_sum_assert(subseq[0:i+1], '>', 0) + "\n"
			assert i != len(subseq) - 1		# not the last one
			# enter idle period. refcount must == 0
		elif i == len(subseq) - 1:
			out_exp += subseq_sum_assert(subseq[0:i+1], '=', 0) + "\n"
	return out_exp

def add_exp(exp, var):
	return "(+ %s %s)" %(exp, var)

def process_subseq2(subseq):
	sys.stderr.write("processing busy periods (%d records)\n" %len(subseq))
	sum_exp = ""
	out_exp = ""
	for i in xrange(0, len(subseq)):
		var = subseq[i].replace("\n", "")
		if i == 0:
			sum_exp = var
		else:
			sum_exp = add_exp(sum_exp, var)
		if is_var_bb(var) and is_var_get(var):
			# a new basic block. refcount must > 0 
			out_exp += "(assert(%s %s %d))" %(">", sum_exp, 0) + "\n"
			assert i != len(subseq) - 1		# not the last one
			# enter idle period. refcount must == 0
		elif i == len(subseq) - 1:
			out_exp += "(assert(%s %s %d))" %("=", sum_exp, 0) + "\n"
	return out_exp
		
all_vars = []	# non unique, may contain "I"
all_get = []	# unique, does not contain "I"
all_put = []	# unique, does not contain "I"
all_get_vars = {}
all_put_vars = {}
idle_period_index = []

def init_process(input_fname):
	global all_get_vars
	global all_put_vars
	global idle_period_index
	global all_vars
	
	f = file(input_fname)
	all_vars = f.readlines()
	for i, line in enumerate(all_vars):
		line = line.replace("\n", "")
		if (len(line) == 0):		# empty line
			continue
		if is_var_idle(line):		# remember idle index so we can skip later
			idle_period_index.append(i)
		if is_var_get(line) and not line in all_get_vars:
			all_get_vars[line] = ""
			all_get.append(line)
		if is_var_put(line) and not line in all_put_vars:
			all_put_vars[line] = ""
			all_put.append(line)

def gen_decls():
	out_exp = ""
	
	for k, v in all_get_vars.iteritems():
		out_exp += "(declare-const %s Int)\n" %k
	for k, v in all_put_vars.iteritems():
		out_exp += "(declare-const %s Int)\n" %k

	for k, v in all_get_vars.iteritems():
		out_exp += "(assert (or (= %s 0) (= %s 1)))\n" %(k, k)
	for k, v in all_put_vars.iteritems():
		out_exp += "(assert (or (= %s 0) (= %s -1)))\n" %(k, k)
	
	return out_exp

def gen_constraints():
	start = idle_period_index[0] + 1
	for i in idle_period_index[1:]:
		print "; constraint from idle period (records %d-%d)" %(start, i)
		assert start < i
		print process_subseq2(all_vars[start:i])
		start = i + 1  # skip the idle 
		if (start >= len (all_vars)):
			break

def gen_min_gpcount(gcount, pcount):
	global all_get
	global all_put
	print "; as specified, total #get %d, #put %d" %(gcount, pcount)
	print subseq_sum_assert(all_get, "<=", gcount)
	print subseq_sum_assert(all_put, ">=", -pcount)

def gen_footer():
	print ''' 
(check-sat)
(get-model)'''
	
if __name__ == "__main__":
	if (len(sys.argv) < 3):
		usage()
		sys.exit(1)
		
	#sys.setrecursionlimit(20000)

	init_process(sys.argv[1])
	sys.stderr.write("init processing done\n")
	#print idle_period_index
	print gen_decls()
	sys.stderr.write("start gen constraints...\n")
	gen_constraints()
	gen_min_gpcount(int(sys.argv[2]), int(sys.argv[3]))
	gen_footer()
	
